/*
    Sega Saturn memory map viewer
    by Charles MacDonald
    WWW: http://cgfm2.emuviews.com

    Based on the bmp8bpp.zip and startup.zip packages written by
    Bart Trzynadlowski. (www.dynarec.com/~bart)

    Uses the SH-COFF toolchain from KPIT Infosystems. (www.kpit.com)
    Load and run at address $06004000.

    Controls:

    Left/Right  - Change digit position
    Up/Down     - Change digit value
    A           - Re-read memory
    B           - Cycle through predefined locations
    C           - Increment address by 512K
    Start       - Return to cheat cartridge menu

    Bit 0 of the address is ignored, you can only read from even addresses.
    Setting bit 0 will byte-swap the data read at the same address.
*/

#include "shared.h"

uint32 def_addr[16] =
{
    0x20000000, /* BIOS */
    0x20100000, /* SMPC Registers */
    0x20180000, /* Backup RAM */
    0x20200000, /* Work RAM (Low) */
    0x22000000, /* Action Replay ROM */
    0x22400000, /* Action Replay RAM (1st 512K) */
    0x22600000, /* Action Replay RAM (2nd 512K) */
    0x25800000, /* CD-ROM Registers */
    0x25C00000, /* VDP1 VRAM */
    0x25C80000, /* VDP1 Framebuffer */
    0x25D00000, /* VDP1 Registers */
    0x25E00000, /* VDP2 VRAM */
    0x25F00000, /* VDP2 CRAM */
    0x25F80000, /* VDP2 Registers */
    0x25FE0000, /* SCU Registers */
    0x26000000, /* Work RAM (High) */
};

void dump_mem(uint32 address)
{
    int x, y;
    vuint8 *memory = (vuint8 *)0;
    vuint16 *ptr = (vuint16 *)&memory[address & 0xFFFFFFFE];

    for(y = 0; y < 16; y++)
    for(x = 0; x < 8; x++)
    {
        uint16 temp = *ptr++;
        if(address & 1)
            temp = (temp >> 8) | (temp << 8);

        _printhexb( 3+x*6+0, 7+y, 0xF1, temp >> 8);
        _printhexb( 3+x*6+3, 7+y, 0x71, temp & 0xFF);
        _printchar(53+x*2+0, 7+y, 0xE1, temp >> 8);
        _printchar(53+x*2+1, 7+y, 0xE1, temp & 0xFF);
    }
}


void _main()
{
    int x, y;
    int location = 0;
    int update = 1;     /* Memory dump update flag */
    int position = 7;   /* Digit position in address */
    int cursorx = 3;  /* Cursor position */
    int cursory = 5;

    uint32 address = def_addr[4];
    uint16 joy_old, joy_new, joy_delta;

    vdp_init();
    conio_init();
    pad_init();

    /* Clear background */
    for(y = 0; y < 16; y++)
    for(x = 0; x < 48; x++)
        _printchar(3+x, 7+y, 0x01, ' ');

    /* Print messages and cursor */
    _printf(3, 1, 0xF0, "Sega Saturn Memory Map Viewer");
    _printf(3, 2, 0xF0, "by Charles MacDonald - http://cgfm2.emuviews.com");

    cursorx = (3+position);
    _printchar(cursorx, cursory, 0x70, 0x7F);

    /* Screen on */
    TVMD = 0x8002;

    for(;;)
    {
        wait_vblank_out();
        wait_vblank_in();

        /* Poll joypad */
        joy_old = joy_new;
        joy_new = pad_read(0);
        joy_delta = (joy_new ^ joy_old) & joy_new;

        /* Start - Quit program */
        if(joy_new & PAD_START)
            break;

        /* Left/Right - Change digit position */
        if(joy_delta & (PAD_LEFT | PAD_RIGHT))
        {
            if(joy_delta & PAD_LEFT)
                position = (position - 1) & 7;
            else
                position = (position + 1) & 7;

            _printchar(cursorx, cursory, 0x70, 0x00);
            cursorx = 3 + position;
            _printchar(cursorx, cursory, 0x70, 0x7F);
        }

        /* Up/Down - Change digit */
        if(joy_delta & (PAD_UP | PAD_DOWN))
        {
            uint32 shift = (position ^ 7) << 2;
            uint32 digit;

            digit = (address >> shift) & 0x0F;
                      
            if(joy_delta & PAD_UP)
                digit = (digit + 1) & 0x0F;
            else
                digit = (digit - 1) & 0x0F;

            address &= (uint32)~(0x0F << shift);
            address |= (digit << shift);
            update = 1;
        }

        /* Button A - Refresh screen */
        if(joy_delta & PAD_A)
            update = 1;

        /* Button B - Cycle through defined addresses */
        if(joy_delta & PAD_B)
        {
            location = (location + 1) & 0x0F;
            address = def_addr[location];
            update = 1;
        }

        /* Button C - Up 512K */
        if(joy_delta & PAD_C)
        {
            address += 0x80000;
            address &= 0xFFFFFFFE;
            update = 1;
        }

        /* Print source address */
        _printhexl(3, 4, 0x71, address);
        
        if(update)
        {
            update = 0;
            TVMD = 0x0002;
            dump_mem(address);
            TVMD = 0x8002;
        }
    }
}

